﻿using Humanizer;
using ShellProgressBar;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Drawing;
using System.Globalization;
using System.Text.Json;
using View_by_Distance.Shared.Models;
using View_by_Distance.Shared.Models.Stateless.Methods;
using WindowsShortcutFactory;

namespace View_by_Distance.Map.Models.Stateless;

internal abstract class MapLogic
{

    internal record Record(string PersonKeyFormatted,
                           int DirectoryNumber,
                           string PersonDisplayDirectoryName,
                           bool IsDefault,
                           string MappedFaceFile);

    internal record TicksDirectory(string Directory,
                                   string DirectoryName,
                                   DateTime DirectoryDateTime,
                                   DateTime AlternateDirectoryDateTime,
                                   float? TotalDays);

    internal record PersonKeyFormattedIdThenWholePercentages(string PersonKeyFormatted,
                                                             string PersonDisplayDirectoryName,
                                                             bool IsDefault,
                                                             string MappedFaceFile,
                                                             int Id,
                                                             int WholePercentages);

    internal static void SetSkipCollections(Configuration configuration, ReadOnlyCollection<PersonContainer> personContainers, string? a2PeopleSingletonDirectory, Dictionary<int, List<(string, int)>> skipCollection, Dictionary<int, List<(string, int)>> skipNotSkipCollection)
    {
        int? id;
        string fileName;
        int? wholePercentages;
        List<string> distinctFiles = new();
        List<string> distinctFileName = new();
        bool skipNotSkipDirectoriesAny = configuration.SkipNotSkipDirectories.Length > 0;
        string[] checkDirectories = (from l in configuration.SkipNotSkipDirectories select Path.GetFullPath($"{a2PeopleSingletonDirectory}{l}")).ToArray();
        foreach (PersonContainer personContainer in personContainers)
        {
            foreach (string personDisplayDirectoryAllFile in personContainer.DisplayDirectoryAllFiles)
            {
                if (!personDisplayDirectoryAllFile.EndsWith(configuration.FacesFileNameExtension))
                    continue;
                if (distinctFiles.Contains(personDisplayDirectoryAllFile))
                    continue;
                distinctFiles.Add(personDisplayDirectoryAllFile);
            }
        }
        foreach (string distinctFile in distinctFiles)
        {
            fileName = Path.GetFileName(distinctFile);
            if (distinctFileName.Contains(fileName))
            {
                if (!distinctFile.EndsWith(".dup") && !File.Exists($"{distinctFile}.dup"))
                    File.Move(distinctFile, $"{distinctFile}.dup");
                continue;
            }
            (id, wholePercentages) = IMapping.GetConverted(configuration.FacesFileNameExtension, distinctFile);
            if (id is null || wholePercentages is null)
                continue;
            if (!skipNotSkipDirectoriesAny || !checkDirectories.Any(l => distinctFile.StartsWith(l)))
            {
                if (!skipCollection.ContainsKey(id.Value))
                    skipCollection.Add(id.Value, new());
                skipCollection[id.Value].Add((distinctFile, wholePercentages.Value));
            }
            else
            {
                if (!skipNotSkipCollection.ContainsKey(id.Value))
                    skipNotSkipCollection.Add(id.Value, new());
                skipNotSkipCollection[id.Value].Add((distinctFile, wholePercentages.Value));
            }
        }
    }

    internal static void SetPersonCollections(Configuration configuration, ReadOnlyCollection<PersonContainer> personContainers, Dictionary<string, string> personKeyFormattedToNewestPersonKeyFormatted, List<string> personKeyFormattedCollection)
    {
        string personKeyFormatted;
        string newestPersonKeyFormatted;
        foreach (PersonContainer personContainer in personContainers)
        {
            if (personContainer.Key is null || personContainer.Birthdays is null || personContainer.Birthdays.Length == 0)
                continue;
            foreach (PersonBirthday personBirthday in personContainer.Birthdays)
            {
                personKeyFormatted = IPersonBirthday.GetFormatted(configuration.PersonBirthdayFormat, personBirthday);
                personKeyFormattedCollection.Add(personKeyFormatted);
                if (personContainer.Birthdays.Length < 1)
                    continue;
                newestPersonKeyFormatted = IPersonBirthday.GetFormatted(configuration.PersonBirthdayFormat, personContainer.Key.Value);
                if (!personKeyFormattedToNewestPersonKeyFormatted.ContainsKey(personKeyFormatted))
                    personKeyFormattedToNewestPersonKeyFormatted.Add(personKeyFormatted, newestPersonKeyFormatted);
            }
        }
    }

    private static void MoveTo(string actionDirectory, TicksDirectory ticksDirectory, string directory, string personKeyFormatted, string yearDirectoryName, string alphaDirectoryName, string[] files, string[] facesFileNames)
    {
        string checkFile;
        string actionDirectoryName = Path.GetFileName(actionDirectory);
        string checkDirectory = actionDirectoryName.StartsWith("y", StringComparison.CurrentCultureIgnoreCase) ? Path.Combine(ticksDirectory.Directory, personKeyFormatted, yearDirectoryName, alphaDirectoryName) : Path.Combine(directory, actionDirectoryName);
        if (!Directory.Exists(checkDirectory))
            _ = Directory.CreateDirectory(checkDirectory);
        foreach (string file in files)
        {
            if (facesFileNames.Contains(file))
            {
                checkFile = Path.Combine(checkDirectory, Path.GetFileName(file));
                if (File.Exists(checkFile))
                    continue;
                File.Move(file, checkFile);
                continue;
            }
            File.Delete(file);
        }
    }

    private static void Individually(Configuration configuration, TicksDirectory ticksDirectory, string directory)
    {
        bool isDefault;
        string[] files;
        FileInfo[] collection;
        string[] facesFileNames;
        string yearDirectoryName;
        string[] yearDirectories;
        string alphaDirectoryName;
        string matchDirectoryName;
        string personKeyFormatted;
        string[] alphaDirectories;
        string[] matchDirectories;
        string[] actionDirectories;
        string personDisplayDirectory;
        string[] personKeyDirectories;
        string[] segmentCDirectories = Directory.GetDirectories(directory, "*", SearchOption.TopDirectoryOnly);
        foreach (string segmentCDirectory in segmentCDirectories)
        {
            personKeyDirectories = Directory.GetDirectories(segmentCDirectory, "*", SearchOption.TopDirectoryOnly);
            foreach (string personKeyDirectory in personKeyDirectories)
            {
                personKeyFormatted = Path.GetFileName(personKeyDirectory);
                yearDirectories = Directory.GetDirectories(personKeyDirectory, "*", SearchOption.TopDirectoryOnly);
                foreach (string yearDirectory in yearDirectories)
                {
                    yearDirectoryName = Path.GetFileName(yearDirectory);
                    if (yearDirectoryName.StartsWith('='))
                        Directory.Move(yearDirectory, yearDirectory.Replace('=', '~'));
                }
                yearDirectories = Directory.GetDirectories(personKeyDirectory, "*", SearchOption.TopDirectoryOnly);
                foreach (string yearDirectory in yearDirectories)
                {
                    yearDirectoryName = Path.GetFileName(yearDirectory);
                    matchDirectories = Directory.GetDirectories(yearDirectory, "*", SearchOption.TopDirectoryOnly);
                    alphaDirectories = matchDirectories.Where(l => !long.TryParse(Path.GetFileName(l), out long a)).ToArray();
                    if (alphaDirectories.Length == 0)
                        continue;
                    alphaDirectoryName = Path.GetFileName(alphaDirectories[0]);
                    foreach (string matchDirectory in matchDirectories)
                    {
                        matchDirectoryName = Path.GetFileName(matchDirectory);
                        files = Directory.GetFiles(matchDirectory, "*", SearchOption.TopDirectoryOnly);
                        if (files.Length != 4)
                            continue;
                        collection = files.Select(l => new FileInfo(l)).ToArray();
                        isDefault = IPerson.IsDefaultName(alphaDirectoryName) && IPersonBirthday.IsCounterPersonYear(personKeyFormatted[..4]);
                        if (isDefault)
                            facesFileNames = (from l in collection where l.Extension == configuration.FacesFileNameExtension select l.FullName).ToArray();
                        else
                            facesFileNames = (from l in collection where l.Extension == configuration.FacesFileNameExtension && l.Name.Contains(matchDirectoryName) select l.FullName).ToArray();
                        if (facesFileNames.Length == 0)
                            continue;
                        personDisplayDirectory = Path.Combine(matchDirectory, alphaDirectoryName);
                        if (!Directory.Exists(personDisplayDirectory) || !Directory.Exists(matchDirectory))
                            continue;
                        _ = Process.Start("explorer", matchDirectory);
                        for (int i = 0; i < int.MaxValue; i++)
                        {
                            Thread.Sleep(500);
                            actionDirectories = Directory.GetDirectories(matchDirectory, "*", SearchOption.TopDirectoryOnly).Where(l => l != personDisplayDirectory && !l.EndsWith("Maybe")).ToArray();
                            if (actionDirectories.Length > 0)
                            {
                                MoveTo(actionDirectories[0], ticksDirectory, directory, personKeyFormatted, yearDirectoryName, alphaDirectoryName, files, facesFileNames);
                                break;
                            }
                        }
                    }
                }
            }
        }
    }

    private static void MoveFiles(string personKeyFormatted, string personKeyDirectory, string newestPersonKeyFormatted, string newestPersonKeyDirectory)
    {
        string[] files;
        string checkFile;
        string? checkDirectory;
        string[] directories = Directory.GetDirectories(personKeyDirectory, "*", SearchOption.TopDirectoryOnly);
        foreach (string directory in directories)
        {
            checkDirectory = Path.Combine(newestPersonKeyDirectory, Path.GetFileName(directory));
            if (!Directory.Exists(checkDirectory))
                Directory.Move(directory, checkDirectory);
            else
            {
                files = Directory.GetFiles(directory, "*", SearchOption.AllDirectories);
                foreach (string file in files)
                {
                    if (file.Split(personKeyFormatted).Length != 2 || file.Contains(newestPersonKeyFormatted))
                        continue;
                    checkFile = file.Replace(personKeyFormatted, newestPersonKeyFormatted);
                    checkDirectory = Path.GetDirectoryName(checkFile);
                    if (checkDirectory is null)
                        continue;
                    if (File.Exists(checkFile))
                        continue;
                    if (!Directory.Exists(checkDirectory))
                        _ = Directory.CreateDirectory(checkDirectory);
                    File.Move(file, checkFile);
                }
            }
        }
    }

    private static void MovedToNewestPersonKeyFormatted(string personKeyFormatted, string newestPersonKeyFormatted, TicksDirectory ticksDirectory, string personKeyDirectory)
    {
        string newestPersonKeyDirectory = Path.Combine(ticksDirectory.Directory, newestPersonKeyFormatted);
        if (Directory.Exists(newestPersonKeyDirectory))
            MoveFiles(personKeyFormatted, personKeyDirectory, newestPersonKeyFormatted, newestPersonKeyDirectory);
        else
            Directory.Move(personKeyDirectory, newestPersonKeyDirectory);
    }

    private static void RenameUnknown(string[] files)
    {
        foreach (string file in files)
        {
            if (file.EndsWith(".unk"))
                continue;
            File.Move(file, $"{file}.unk");
        }
    }

    internal static List<Record> DeleteEmptyDirectoriesAndGetCollection(Configuration configuration, long ticks, string eDistanceContentDirectory, ReadOnlyDictionary<string, string> personKeyFormattedToNewestPersonKeyFormatted, ReadOnlyCollection<string> personKeyFormattedCollection)
    {
        List<Record> results = new();
        int? id;
        bool check;
        bool isDefault;
        string message;
        string[] files;
        string fileName;
        int totalSeconds;
        DateTime dateTime;
        TimeSpan timeSpan;
        int directoryNumber;
        int? wholePercentages;
        string? checkDirectory;
        ProgressBar progressBar;
        string[] yearDirectories;
        string personKeyFormatted;
        string? personFirstInitial;
        bool isReservedDirectoryName;
        List<string> distinct = new();
        string[] personNameDirectories;
        string? newestPersonKeyFormatted;
        string personDisplayDirectoryName;
        string[] personNameLinkDirectories;
        string? personFirstInitialDirectory;
        List<TicksDirectory> ticksDirectories;
        string[] personKeyFormattedDirectories;
        string manualCopyHumanized = nameof(Shared.Models.Stateless.IMapLogic.ManualCopy).Humanize(LetterCasing.Title);
        string forceSingleImageHumanized = nameof(Shared.Models.Stateless.IMapLogic.ForceSingleImage).Humanize(LetterCasing.Title);
        ProgressBarOptions options = new() { ProgressCharacter = '─', ProgressBarOnBottom = true, DisableBottomPercentage = true };
        for (int i = 1; i < 3; i++)
        {
            check = false;
            results.Clear();
            distinct.Clear();
            directoryNumber = 0;
            ticksDirectories = UpdateDateVerifyAndGetTicksDirectories(eDistanceContentDirectory);
            totalSeconds = (int)Math.Floor(new TimeSpan(DateTime.Now.Ticks - ticks).TotalSeconds);
            message = $"{i}) {ticksDirectories.Count:000} compile from and clean ticks Director(ies) - B - {totalSeconds} total second(s)";
            progressBar = new(ticksDirectories.Count, message, options);
            foreach (TicksDirectory ticksDirectory in ticksDirectories)
            {
                if (i == 1)
                    progressBar.Tick();
                personKeyFormattedDirectories = Directory.GetDirectories(ticksDirectory.Directory, "*", SearchOption.TopDirectoryOnly);
                foreach (string personKeyFormattedDirectory in personKeyFormattedDirectories)
                {
                    personKeyFormatted = Path.GetFileName(personKeyFormattedDirectory);
                    isReservedDirectoryName = personKeyFormatted.StartsWith(nameof(Shared.Models.Stateless.IMapLogic.Sorting)) || personKeyFormatted.StartsWith(nameof(Shared.Models.Stateless.IMapLogic.Mapping)) || personKeyFormatted.StartsWith(nameof(Shared.Models.Stateless.IMapLogic.ManualCopy));
                    if (!isReservedDirectoryName && personKeyFormatted.StartsWith(nameof(Shared.Models.Stateless.IMapLogic.Individually)))
                    {
                        Individually(configuration, ticksDirectory, personKeyFormattedDirectory);
                        throw new Exception($"B) Move personKey directories up one from {nameof(Shared.Models.Stateless.IMapLogic.Sorting)} and delete {nameof(Shared.Models.Stateless.IMapLogic.Sorting)} directory!");
                    }
                    _ = personKeyFormattedToNewestPersonKeyFormatted.TryGetValue(personKeyFormatted, out newestPersonKeyFormatted);
                    if (personKeyFormattedToNewestPersonKeyFormatted.Count > 0 && newestPersonKeyFormatted is null)
                    {
                        timeSpan = new TimeSpan(DateTime.Now.Ticks - ticksDirectory.DirectoryDateTime.Ticks);
                        if (timeSpan.TotalDays > 6)
                            throw new Exception($"{configuration.MappingDefaultName} are only allowed within x days!");
                    }
                    yearDirectories = Directory.GetDirectories(personKeyFormattedDirectory, "*", SearchOption.TopDirectoryOnly);
                    foreach (string yearDirectory in yearDirectories)
                    {
                        if (check && !Directory.Exists(yearDirectory))
                            continue;
                        files = Directory.GetFiles(yearDirectory, "*", SearchOption.TopDirectoryOnly);
                        personNameDirectories = Directory.GetDirectories(yearDirectory, "*", SearchOption.TopDirectoryOnly);
                        foreach (string file in files)
                            File.Delete(file);
                        foreach (string personNameDirectory in personNameDirectories)
                        {
                            directoryNumber++;
                            personDisplayDirectoryName = Path.GetFileName(personNameDirectory);
                            isDefault = IPerson.IsDefaultName(personDisplayDirectoryName) && IPersonBirthday.IsCounterPersonYear(personKeyFormatted[..4]);
                            if (isDefault && personDisplayDirectoryName.Length == 1)
                            {
                                if (personKeyFormatted.Length != configuration.PersonBirthdayFormat.Length || !DateTime.TryParseExact(personKeyFormatted, configuration.PersonBirthdayFormat, CultureInfo.InvariantCulture, DateTimeStyles.None, out dateTime))
                                    continue;
                                checkDirectory = Path.Combine(yearDirectory, $"X]{dateTime.Ticks}");
                                if (Directory.Exists(checkDirectory))
                                {
                                    Directory.Delete(yearDirectory, recursive: true);
                                    continue;
                                }
                                Directory.Move(personNameDirectory, checkDirectory);
                                if (!check)
                                    check = true;
                                continue;
                            }
                            if (isDefault && (ticksDirectory.DirectoryDateTime.Hour != 0 || ticksDirectory.DirectoryDateTime.Minute != 0 || ticksDirectory.DirectoryDateTime.Second != 0))
                            {
                                checkDirectory = Path.GetDirectoryName(ticksDirectory.Directory);
                                if (checkDirectory is null)
                                    continue;
                                checkDirectory = Path.Combine(checkDirectory, $"({ticksDirectory.AlternateDirectoryDateTime.Ticks})");
                                if (!Directory.Exists(checkDirectory))
                                    _ = Directory.CreateDirectory(checkDirectory);
                                checkDirectory = Path.Combine(checkDirectory, personKeyFormatted);
                                if (!Directory.Exists(checkDirectory))
                                {
                                    Directory.Move(personKeyFormattedDirectory, checkDirectory);
                                    if (!check)
                                        check = true;
                                    break;
                                }
                            }
                            files = Directory.GetFiles(personNameDirectory, "*", SearchOption.TopDirectoryOnly);
                            if (isReservedDirectoryName && files.Length > 0)
                                throw new Exception($"Move personKey directories up one from {nameof(Shared.Models.Stateless.IMapLogic.Sorting)} and delete {nameof(Shared.Models.Stateless.IMapLogic.Sorting)} directory!");
                            if (personKeyFormatted == manualCopyHumanized && files.Length > 0)
                                throw new Exception($"Move personKey directories up one from {manualCopyHumanized} and delete {manualCopyHumanized} directory!");
                            if (personKeyFormatted == forceSingleImageHumanized && files.Length > 0)
                                throw new Exception($"Move personKey directories up one from {forceSingleImageHumanized} and delete {forceSingleImageHumanized} directory!");
                            if (!isDefault)
                            {
                                if (personKeyFormattedToNewestPersonKeyFormatted.Count > 0 && newestPersonKeyFormatted is null)
                                    RenameUnknown(files);
                                else if (newestPersonKeyFormatted is not null && personKeyFormatted != newestPersonKeyFormatted)
                                {
                                    if (!check)
                                        check = true;
                                    MovedToNewestPersonKeyFormatted(personKeyFormatted, newestPersonKeyFormatted, ticksDirectory, personKeyFormattedDirectory);
                                    continue;
                                }
                            }
                            if (personKeyFormatted.Length != configuration.PersonBirthdayFormat.Length)
                                continue;
                            if (personDisplayDirectoryName.Length == 1 || isDefault || !personKeyFormattedCollection.Contains(personKeyFormatted))
                                personFirstInitialDirectory = personNameDirectory;
                            else
                            {
                                personFirstInitial = personDisplayDirectoryName[..1];
                                if (personFirstInitial.All(l => char.IsDigit(l)))
                                {
                                    foreach (string file in files)
                                        File.Delete(file);
                                    files = Directory.GetFiles(personNameDirectory, "*", SearchOption.AllDirectories);
                                    foreach (string file in files)
                                        File.Delete(file);
                                    _ = IPath.DeleteEmptyDirectories(personNameDirectory);
                                    continue;
                                }
                                personFirstInitialDirectory = Path.Combine(yearDirectory, personFirstInitial.ToString());
                                if (Directory.Exists(personFirstInitialDirectory))
                                    throw new Exception("Forgot to ...");
                                Directory.Move(personNameDirectory, personFirstInitialDirectory);
                                files = Directory.GetFiles(personFirstInitialDirectory, "*", SearchOption.TopDirectoryOnly);
                            }
                            foreach (string mappedFaceFile in files)
                            {
                                if (mappedFaceFile.EndsWith(".lnk"))
                                    continue;
                                (id, wholePercentages) = IMapping.GetConverted(configuration.FacesFileNameExtension, mappedFaceFile);
                                if (id is null || wholePercentages is null)
                                    continue;
                                fileName = Path.GetFileName(mappedFaceFile);
                                if (distinct.Contains(fileName))
                                {
                                    if (!mappedFaceFile.EndsWith(".dup") && !File.Exists($"{mappedFaceFile}.dup"))
                                        File.Move(mappedFaceFile, $"{mappedFaceFile}.dup");
                                    continue;
                                }
                                distinct.Add(fileName);
                                results.Add(new(personKeyFormatted, directoryNumber, personDisplayDirectoryName, isDefault, mappedFaceFile));
                            }
                            personNameLinkDirectories = Directory.GetDirectories(personFirstInitialDirectory, "*", SearchOption.TopDirectoryOnly);
                            foreach (string personNameLinkDirectory in personNameLinkDirectories)
                            {
                                files = Directory.GetFiles(personNameLinkDirectory, "*", SearchOption.TopDirectoryOnly);
                                foreach (string file in files)
                                {
                                    if (!file.EndsWith(".lnk"))
                                        continue;
                                    File.Delete(file);
                                }
                                _ = IPath.DeleteEmptyDirectories(personNameLinkDirectory);
                            }
                            _ = IPath.DeleteEmptyDirectories(personFirstInitialDirectory);
                        }
                        _ = IPath.DeleteEmptyDirectories(yearDirectory);
                    }
                    _ = IPath.DeleteEmptyDirectories(personKeyFormattedDirectory);
                }
                _ = IPath.DeleteEmptyDirectories(ticksDirectory.Directory);
                _ = IPath.DeleteEmptyDirectories(ticksDirectory.Directory);
            }
            progressBar.Dispose();
            if (check)
                continue;
            break;
        }
        return results;
    }

    internal static void SetKeyValuePairsAndAddToCollections(Configuration configuration, ReadOnlyCollection<PersonContainer> personContainers, Dictionary<long, PersonContainer> personKeyToPersonContainer, ReadOnlyCollection<PersonKeyFormattedIdThenWholePercentages> personKeyFormattedIdThenWholePercentagesCollection, Dictionary<long, int> personKeyToCount, Dictionary<string, PersonContainer> personKeyFormattedToPersonContainer, Dictionary<long, List<PersonContainer>> personKeyToPersonContainerCollection, List<(PersonKeyFormattedIdThenWholePercentages, PersonContainer)> possiblyNewPersonDisplayDirectoryNamesAndPersonContainer)
    {
        PersonBirthday? personBirthday;
        Dictionary<int, Dictionary<int, List<PersonContainer>>> idThenWholePercentagesToPersonContainers = new();
        (long, PersonContainer)[] collection = GetDistinctCollection(configuration, personContainers, personKeyToPersonContainerCollection, personKeyFormattedToPersonContainer);
        foreach ((long personKey, PersonContainer personContainer) in collection)
            personKeyToPersonContainer.Add(personKey, personContainer);
        if (personKeyFormattedIdThenWholePercentagesCollection.Count > 0)
        {
            string group;
            char[] matches;
            char status, sex, first;
            PersonDirectory personDirectory;
            PersonContainer? personContainer;
            foreach (PersonKeyFormattedIdThenWholePercentages personKeyFormattedIdThenWholePercentages in personKeyFormattedIdThenWholePercentagesCollection)
            {
                personBirthday = IPersonBirthday.GetPersonBirthday(configuration.PersonBirthdayFormat, personKeyFormattedIdThenWholePercentages.PersonKeyFormatted);
                if (personBirthday is null)
                    throw new Exception();
                if (!personKeyFormattedToPersonContainer.TryGetValue(personKeyFormattedIdThenWholePercentages.PersonKeyFormatted, out personContainer))
                {
                    matches = configuration.PersonCharacters.Where(l => personKeyFormattedIdThenWholePercentages.PersonDisplayDirectoryName.Contains(l)).ToArray();
                    if (matches.Length == 0)
                        throw new NotSupportedException();
                    group = IPerson.GetHourGroup(personKeyFormattedIdThenWholePercentages.PersonDisplayDirectoryName, personBirthday.Value.Hour);
                    (status, sex, first) = IPerson.GetPersonHour(personKeyFormattedIdThenWholePercentages.PersonDisplayDirectoryName, personBirthday.Value.Hour);
                    personDirectory = new(matches.First(), group, status, sex, first);
                    personContainer = new(configuration.PersonCharacters.ToArray(), personBirthday, personKeyFormattedIdThenWholePercentages.PersonDisplayDirectoryName, personDirectory);
                    personKeyFormattedToPersonContainer.Add(personKeyFormattedIdThenWholePercentages.PersonKeyFormatted, personContainer);
                }
                if (personContainer.Key is null)
                    throw new Exception();
                if (!personKeyToCount.ContainsKey(personContainer.Key.Value))
                    personKeyToCount.Add(personContainer.Key.Value, 0);
                personKeyToCount[personContainer.Key.Value]++;
                possiblyNewPersonDisplayDirectoryNamesAndPersonContainer.Add(new(personKeyFormattedIdThenWholePercentages, personContainer));
            }
        }
    }

    private static Dictionary<int, Dictionary<int, List<PersonContainer>>> GetAll(Configuration configuration, ReadOnlyDictionary<string, PersonContainer> personKeyFormattedToPersonContainer, ReadOnlyCollection<PersonKeyFormattedIdThenWholePercentages> personKeyFormattedIdThenWholePercentagesCollection)
    {
        Dictionary<int, Dictionary<int, List<PersonContainer>>> results = new();
        PersonBirthday? personBirthday;
        PersonContainer? personContainer;
        List<PersonContainer>? personContainers;
        Dictionary<int, List<PersonContainer>>? idTo;
        if (personKeyFormattedIdThenWholePercentagesCollection.Count > 0)
        {
            foreach (PersonKeyFormattedIdThenWholePercentages personKeyFormattedIdThenWholePercentages in personKeyFormattedIdThenWholePercentagesCollection)
            {
                personBirthday = IPersonBirthday.GetPersonBirthday(configuration.PersonBirthdayFormat, personKeyFormattedIdThenWholePercentages.PersonKeyFormatted);
                if (personBirthday is null)
                    throw new Exception();
                if (!personKeyFormattedToPersonContainer.TryGetValue(personKeyFormattedIdThenWholePercentages.PersonKeyFormatted, out personContainer))
                    throw new Exception();
                if (!results.TryGetValue(personKeyFormattedIdThenWholePercentages.Id, out idTo))
                {
                    results.Add(personKeyFormattedIdThenWholePercentages.Id, new());
                    if (!results.TryGetValue(personKeyFormattedIdThenWholePercentages.Id, out idTo))
                        throw new Exception();
                }
                if (!idTo.TryGetValue(personKeyFormattedIdThenWholePercentages.WholePercentages, out personContainers))
                {
                    idTo.Add(personKeyFormattedIdThenWholePercentages.WholePercentages, new());
                    if (!idTo.TryGetValue(personKeyFormattedIdThenWholePercentages.WholePercentages, out personContainers))
                        throw new Exception();
                }
                personContainers.Add(personContainer);
            }
        }
        return results;
    }

    private static ReadOnlyDictionary<int, ReadOnlyDictionary<int, ReadOnlyCollection<PersonContainer>>> GetReadOnly(Dictionary<int, Dictionary<int, List<PersonContainer>>> idThenWholePercentagesToPersonContainerCollection)
    {
        Dictionary<int, ReadOnlyDictionary<int, ReadOnlyCollection<PersonContainer>>> results = new();
        List<long> distinct = new();
        List<PersonContainer> personContainers;
        Dictionary<int, ReadOnlyCollection<PersonContainer>> keyValuePairs;
        foreach (KeyValuePair<int, Dictionary<int, List<PersonContainer>>> idTo in idThenWholePercentagesToPersonContainerCollection)
        {
            keyValuePairs = new();
            foreach (KeyValuePair<int, List<PersonContainer>> wholePercentagesTo in idThenWholePercentagesToPersonContainerCollection[idTo.Key])
            {
                distinct.Clear();
                personContainers = new();
                foreach (PersonContainer personContainer in wholePercentagesTo.Value)
                {
                    if (personContainer.Key is null)
                        throw new Exception();
                    if (distinct.Contains(personContainer.Key.Value))
                        continue;
                    personContainers.Add(personContainer);
                }
                keyValuePairs.Add(wholePercentagesTo.Key, new(personContainers));
            }
            results.Add(idTo.Key, new(keyValuePairs));
        }
        return new(results);
    }

    internal static ReadOnlyDictionary<int, ReadOnlyDictionary<int, ReadOnlyCollection<PersonContainer>>> GetIdThenWholePercentagesToPersonContainers(Configuration configuration, ReadOnlyDictionary<string, PersonContainer> personKeyFormattedToPersonContainer, ReadOnlyCollection<PersonKeyFormattedIdThenWholePercentages> personKeyFormattedIdThenWholePercentagesCollection)
    {
        ReadOnlyDictionary<int, ReadOnlyDictionary<int, ReadOnlyCollection<PersonContainer>>> results;
        Dictionary<int, Dictionary<int, List<PersonContainer>>> idThenWholePercentagesToPersonContainerCollection;
        idThenWholePercentagesToPersonContainerCollection = GetAll(configuration, personKeyFormattedToPersonContainer, personKeyFormattedIdThenWholePercentagesCollection);
        results = GetReadOnly(idThenWholePercentagesToPersonContainerCollection);
        return results;
    }

    private static (long, PersonContainer)[] GetDistinctCollection(Configuration configuration, IEnumerable<PersonContainer> personContainers, Dictionary<long, List<PersonContainer>> personKeyToPersonContainerCollection, Dictionary<string, PersonContainer> personKeyFormattedToPersonContainer)
    {
        (long, PersonContainer)[] results;
        const int zero = 0;
        List<string> errors = new();
        string newestPersonKeyFormatted;
        List<(long PersonKey, PersonContainer PersonContainer)> collection = new();
        foreach (PersonContainer personContainer in personContainers)
        {
            if (personContainer.Key is null)
                continue;
            if (!personKeyToPersonContainerCollection.ContainsKey(personContainer.Key.Value))
                personKeyToPersonContainerCollection.Add(personContainer.Key.Value, new());
            personKeyToPersonContainerCollection[personContainer.Key.Value].Add(personContainer);
            newestPersonKeyFormatted = IPersonBirthday.GetFormatted(configuration.PersonBirthdayFormat, personContainer.Key.Value);
            if (!personKeyFormattedToPersonContainer.ContainsKey(newestPersonKeyFormatted))
                personKeyFormattedToPersonContainer.Add(newestPersonKeyFormatted, personContainer);
        }
        foreach (KeyValuePair<long, List<PersonContainer>> keyValuePair in personKeyToPersonContainerCollection)
        {
            if (keyValuePair.Value.Count != 1 && (from l in keyValuePair.Value select l.DisplayDirectoryName).Distinct().Count() != 1)
                errors.Add(keyValuePair.Value[zero].DisplayDirectoryName);
            collection.Add(new(keyValuePair.Key, keyValuePair.Value[zero]));
        }
        if (errors.Count > 0)
            throw new Exception(string.Join(Environment.NewLine, errors));
        results = (from l in collection orderby l.PersonKey descending select (l.PersonKey, l.PersonContainer)).ToArray();
        return results;
    }

    internal static ReadOnlyCollection<PersonKeyFormattedIdThenWholePercentages> GetPersonKeyFormattedIdThenWholePercentages(Configuration configuration, long ticks, List<Record> collection)
    {
        List<PersonKeyFormattedIdThenWholePercentages> results = new();
        int? id;
        int? wholePercentages;
        List<int> wholePercentagesCollection;
        Dictionary<int, List<int>> idToWholePercentagesCollection = new();
        int totalSeconds = (int)Math.Floor(new TimeSpan(DateTime.Now.Ticks - ticks).TotalSeconds);
        string message = $") {collection.Count:000} join from ticks Director(ies) - C - {totalSeconds} total second(s)";
        ProgressBarOptions options = new() { ProgressCharacter = '─', ProgressBarOnBottom = true, DisableBottomPercentage = true };
        using ProgressBar progressBar = new(collection.Count, message, options);
        foreach (Record record in collection)
        {
            progressBar.Tick();
            (id, wholePercentages) = IMapping.GetConverted(configuration.FacesFileNameExtension, record.MappedFaceFile);
            if (id is null || wholePercentages is null)
                continue;
            if (!idToWholePercentagesCollection.ContainsKey(id.Value))
                idToWholePercentagesCollection.Add(id.Value, new());
            wholePercentagesCollection = idToWholePercentagesCollection[id.Value];
            wholePercentagesCollection.Add(wholePercentages.Value);
            idToWholePercentagesCollection[id.Value].Add(wholePercentages.Value);
            results.Add(new(record.PersonKeyFormatted, record.PersonDisplayDirectoryName, record.IsDefault, record.MappedFaceFile, id.Value, wholePercentages.Value));
        }
        return new(results);
    }

    private static List<PersonContainer> GetNonSpecificPeopleCollection(Configuration configuration, long ticks, List<long> personKeys, ReadOnlyDictionary<long, int> personKeyToCount)
    {
        List<PersonContainer> results = new();
        bool check;
        long personKey;
        int? approximateYears = null;
        PersonBirthday personBirthday;
        PersonContainer personContainer;
        string[] personDisplayDirectoryAllFiles = Array.Empty<string>();
        DateTime incrementDate = new(configuration.PersonBirthdayFirstYear, 1, 1);
        long oneHour = new DateTime(1, 1, 1, 1, 0, 0).Ticks - new DateTime(1, 1, 1).Ticks;
        for (int i = 0; i < int.MaxValue; i++)
        {
            check = false;
            personKey = incrementDate.Ticks;
            incrementDate = incrementDate.AddDays(1);
            if (incrementDate.Ticks > ticks)
                break;
            if (personKeys.Contains(personKey))
                continue;
            if (personKeyToCount.ContainsKey(personKey))
                continue;
            for (int j = 1; j < 24; j++)
            {
                if (personKeys.Contains(personKey + (oneHour * j)))
                {
                    check = true;
                    break;
                }
                if (personKeyToCount.ContainsKey(personKey + (oneHour * j)))
                {
                    check = true;
                    break;
                }
            }
            if (check)
                continue;
            personBirthday = IPersonBirthday.GetPersonBirthday(personKey + (oneHour * 2));
            personContainer = new(approximateYears, new PersonBirthday[] { personBirthday }, personDisplayDirectoryAllFiles, configuration.MappingDefaultName, personKey);
            results.Add(personContainer);
            if (results.Count > 99)
                break;
        }
        return results;
    }

    internal static List<PersonContainer> GetNotMappedPersonContainers(Configuration configuration, long ticks, ReadOnlyCollection<PersonContainer> personContainers, ReadOnlyDictionary<long, int> personKeyToCount)
    {
        List<PersonContainer> results = new();
        List<PersonContainer> notMappedAndNotNamedPersonKeys = new();
        List<PersonContainer> notMappedAndWithNamedPersonKeys = new();
        List<long> personKeys = IPersonContainer.GetPersonKeys(personContainers);
        foreach (PersonContainer personContainer in personContainers)
        {
            if (personContainer.Key is null || personContainer.Birthdays is null || personContainer.Birthdays.Length == 0)
                continue;
            if (personKeys.Contains(personContainer.Key.Value))
                continue;
            if (personKeyToCount.ContainsKey(personContainer.Key.Value))
                continue;
            if (string.IsNullOrEmpty(personContainer.DisplayDirectoryName) || personContainer.DisplayDirectoryName == configuration.MappingDefaultName)
                notMappedAndNotNamedPersonKeys.Add(personContainer);
            else
                notMappedAndWithNamedPersonKeys.Add(personContainer);
        }
        results.AddRange(notMappedAndNotNamedPersonKeys);
        if (results.Count == 0)
            results.AddRange(GetNonSpecificPeopleCollection(configuration, ticks, personKeys, personKeyToCount));
        return results;
    }

    private static string? GetDisplayDirectoryName(Dictionary<long, PersonContainer> personKeyToPersonContainer, long key)
    {
        string? result = null;
        if (personKeyToPersonContainer.TryGetValue(key, out PersonContainer? personContainer))
        {
            result = personContainer.DisplayDirectoryName;
            if (string.IsNullOrEmpty(result))
                throw new NotSupportedException();
        }
        return result;
    }

    internal static void SetPersonKeyToPersonContainer(Configuration configuration, ReadOnlyCollection<PersonContainer> personContainers, ReadOnlyDictionary<long, int> personKeyToCount, Dictionary<long, PersonContainer> personKeyToPersonContainer, ReadOnlyDictionary<long, List<PersonContainer>> personKeyToPersonContainerCollection)
    {
        string? displayDirectoryName;
        foreach (PersonContainer personContainer in personContainers)
        {
            if (personContainer.Key is null || !personKeyToCount.ContainsKey(personContainer.Key.Value))
                continue;
            displayDirectoryName = GetDisplayDirectoryName(personKeyToPersonContainer, personContainer.Key.Value);
            if (displayDirectoryName is not null && (displayDirectoryName == personContainer.DisplayDirectoryName || (displayDirectoryName[0] == personContainer.DisplayDirectoryName[0] && (displayDirectoryName.Length == 1 || personContainer.DisplayDirectoryName.Length == 1))))
                continue;
            personKeyToPersonContainer.Add(personContainer.Key.Value, personContainer);
        }
        if (personKeyToCount.Count > 0)
        {
            const int zero = 0;
            int? approximateYears = null;
            PersonBirthday? personBirthday;
            PersonContainer personContainer;
            displayDirectoryName = configuration.MappingDefaultName;
            foreach (KeyValuePair<long, int> keyValuePair in personKeyToCount)
            {
                if (personKeyToPersonContainer.ContainsKey(keyValuePair.Key))
                    continue;
                personBirthday = IPersonBirthday.GetPersonBirthday(keyValuePair.Key);
                if (!personKeyToPersonContainerCollection.ContainsKey(keyValuePair.Key))
                    personContainer = new(approximateYears, personBirthday, displayDirectoryName, keyValuePair.Key);
                else
                    personContainer = new(approximateYears, personBirthday, personKeyToPersonContainerCollection[keyValuePair.Key][zero].PersonDirectory, displayDirectoryName, keyValuePair.Key);
                personKeyToPersonContainer.Add(keyValuePair.Key, personContainer);
            }
        }
    }

    internal static void PossiblyRebuildPersonContainers(Configuration configuration, long ticks, string? a2PeopleSingletonDirectory, ReadOnlyDictionary<long, int> readOnlyPersonKeyToCount, ReadOnlyCollection<(PersonKeyFormattedIdThenWholePercentages, PersonContainer)> possiblyNewPersonDisplayDirectoryNamesAndPersonContainer)
    {
        int count;
        long personKey;
        bool[] matches;
        string fileName;
        string checkFile;
        const int zero = 0;
        string personKeyFormatted;
        string[] deleteCollection;
        List<string> distinct = new();
        PersonBirthday personBirthday;
        string personDisplayDirectory;
        DateTime dateTime = new(ticks);
        string personKeyFormattedDirectory;
        foreach ((PersonKeyFormattedIdThenWholePercentages personKeyFormattedIdThenWholePercentages, PersonContainer personContainer) in possiblyNewPersonDisplayDirectoryNamesAndPersonContainer)
        {
            if (distinct.Contains(personKeyFormattedIdThenWholePercentages.PersonKeyFormatted))
                continue;
            if (a2PeopleSingletonDirectory is null || personContainer.Key is null || personContainer.Birthdays is null || personContainer.PersonDirectory is null || personContainer.Birthdays.Length == 0)
                continue;
            fileName = $"{Path.GetFileName(personKeyFormattedIdThenWholePercentages.MappedFaceFile)}{configuration.FacesHiddenFileNameExtension}";
            personBirthday = personContainer.Birthdays[zero];
            personKey = personBirthday.Value.Ticks;
            personKeyFormatted = IPersonBirthday.GetFormatted(configuration.PersonBirthdayFormat, personBirthday);
            personDisplayDirectory = Path.Combine(a2PeopleSingletonDirectory, personContainer.PersonDirectory.Char.ToString(), personContainer.PersonDirectory.Group, personContainer.DisplayDirectoryName);
            personKeyFormattedDirectory = Path.GetFullPath(Path.Combine(personDisplayDirectory, personKeyFormatted));
            deleteCollection = (from l in personContainer.DisplayDirectoryAllFiles where l.StartsWith(personKeyFormattedDirectory) select l).ToArray();
            if (personContainer.DisplayDirectoryAllFiles.Length != 0 && deleteCollection.Length == 0)
                throw new NotSupportedException();
            if (!Directory.Exists(personKeyFormattedDirectory))
                _ = Directory.CreateDirectory(personKeyFormattedDirectory);
            _ = readOnlyPersonKeyToCount.TryGetValue(personKey, out count);
            _ = Directory.CreateDirectory(Path.Combine(personDisplayDirectory, count.ToString("0000")));
            Directory.SetLastWriteTime(personDisplayDirectory, dateTime.AddMinutes(count));
            matches = (from l in personContainer.DisplayDirectoryAllFiles where l.EndsWith(fileName) select true).ToArray();
            if (matches.Length > 0)
                continue;
            matches = (from l in personContainer.DisplayDirectoryAllFiles where l.EndsWith(configuration.FacesHiddenFileNameExtension) select true).ToArray();
            if (matches.Length > 0)
                continue;
            if (!File.Exists(personKeyFormattedIdThenWholePercentages.MappedFaceFile))
                continue;
            checkFile = Path.Combine(personKeyFormattedDirectory, $"{Path.GetFileName(personKeyFormattedIdThenWholePercentages.MappedFaceFile)}{configuration.FacesHiddenFileNameExtension}");
            if (File.Exists(checkFile))
                continue;
            File.Copy(personKeyFormattedIdThenWholePercentages.MappedFaceFile, checkFile);
            foreach (string delete in deleteCollection)
            {
                if (delete.EndsWith(".lnk"))
                    continue;
                if (!File.Exists(delete))
                    continue;
                File.Delete(delete);
            }
            Directory.SetLastWriteTime(personDisplayDirectory, DateTime.Now);
            distinct.Add(personKeyFormattedIdThenWholePercentages.PersonKeyFormatted);
        }
    }

    private static List<TicksDirectory> UpdateDateVerifyAndGetTicksDirectories(string eDistanceContentDirectory)
    {
        List<TicksDirectory> results = new();
        float? totalDays;
        string ticksDirectoryName;
        DateTime directoryDateTime;
        DirectoryInfo directoryInfo;
        long? lastDirectoryTicks = null;
        DateTime dateTime = DateTime.Now;
        for (int i = 1; i < 5; i++)
            _ = IPath.DeleteEmptyDirectories(eDistanceContentDirectory);
        if (!Directory.Exists(eDistanceContentDirectory))
            _ = Directory.CreateDirectory(eDistanceContentDirectory);
        string[] ticksDirectories = Directory.GetDirectories(eDistanceContentDirectory, "*", SearchOption.TopDirectoryOnly);
        foreach (string ticksDirectory in ticksDirectories)
        {
            ticksDirectoryName = Path.GetFileName(ticksDirectory);
            if (ticksDirectoryName.Length < 3 || ticksDirectoryName.First() != '(' || ticksDirectoryName[^1] != ')')
                continue;
            if (!long.TryParse(ticksDirectoryName[1..^1], out long directoryTicks))
            {
                if (!long.TryParse(ticksDirectoryName[1..^4], out directoryTicks))
                    throw new NotSupportedException();
            }
            directoryInfo = new(ticksDirectory);
            directoryDateTime = new DateTime(directoryTicks);
            if (directoryInfo.CreationTime.Ticks != directoryTicks)
                Directory.SetCreationTime(ticksDirectory, new DateTime(directoryTicks));
            if (directoryInfo.LastWriteTime.Ticks != directoryTicks)
                Directory.SetLastWriteTime(ticksDirectory, new DateTime(directoryTicks));
            totalDays = lastDirectoryTicks is null || new TimeSpan(dateTime.Ticks - directoryTicks).TotalDays < 1 ? null : (float)new TimeSpan(directoryTicks - lastDirectoryTicks.Value).TotalDays;
            results.Add(new(ticksDirectory, ticksDirectoryName, new(directoryTicks), new DateTime(directoryDateTime.Year, directoryDateTime.Month, directoryDateTime.Day + 1), totalDays));
            lastDirectoryTicks = directoryTicks;
        }
        string[] compare = (from l in results where l.TotalDays is not null and < 3.95f select l.Directory).ToArray();
        if (compare.Length > 0)
            throw new Exception($"Please Consolidate <{string.Join(Environment.NewLine, compare)}>");
        return results;
    }

    private static List<(long, int?, string)> GetDisplayDirectoryAllFiles(string fileNameExtension, ReadOnlyCollection<PersonContainer> personContainers)
    {
        List<(long, int?, string)> results = new();
        string fileName;
        List<string> distinct = new();
        foreach (PersonContainer personContainer in personContainers)
        {
            if (personContainer.Key is null)
                continue;
            for (int i = personContainer.DisplayDirectoryAllFiles.Length - 1; i > -1; i--)
            {
                if (!personContainer.DisplayDirectoryAllFiles[i].EndsWith(fileNameExtension))
                    continue;
                fileName = Path.GetFileName(personContainer.DisplayDirectoryAllFiles[i]);
                if (distinct.Contains(fileName))
                    continue;
                distinct.Add(fileName);
                results.Add(new(personContainer.Key.Value, null, personContainer.DisplayDirectoryAllFiles[i]));
            }
        }
        return results;
    }

    private static List<(long PersonKey, int? DirectoryNumber, string File)> GetCollection(Configuration configuration, ReadOnlyCollection<PersonContainer> personContainers, List<Record> records)
    {
        List<(long PersonKey, int? DirectoryNumber, string File)> results = new();
        string file;
        long personKey;
        string fileName;
        List<string> distinct = new();
        PersonBirthday? personBirthday;
        results.AddRange(GetDisplayDirectoryAllFiles(configuration.FacesFileNameExtension, personContainers));
        foreach (Record record in records)
        {
            personBirthday = IPersonBirthday.GetPersonBirthday(configuration.PersonBirthdayFormat, record.PersonKeyFormatted);
            if (personBirthday is null)
                continue;
            fileName = Path.GetFileName(record.MappedFaceFile);
            if (distinct.Contains(fileName))
                continue;
            distinct.Add(fileName);
            personKey = personBirthday.Value.Ticks;
            results.Add(new(personKey, record.DirectoryNumber, record.MappedFaceFile));
        }
        for (int i = results.Count - 1; i > -1; i--)
        {
            file = results[i].File;
            if (file.EndsWith(".old"))
            {
                results.RemoveAt(i);
                continue;
            }
            if (!file.EndsWith(".dup") && !file.EndsWith(".unk") && !file.EndsWith(".abd"))
                continue;
            if (!File.Exists(file))
                continue;
            File.Move(file, file[..^4]);
            results[i] = new(results[i].PersonKey, results[i].DirectoryNumber, file[..^4]);
        }
        return results;
    }

    private static void ParallelFor(Configuration configuration, string eDistanceContentDirectory, Dictionary<int, List<(string, int)>> skipCollection, List<LocationContainer<MetadataExtractor.Directory>> locationContainers, long personKey, int? directoryNumber, string file)
    {
        string[] fileMatches;
        const string lnk = ".lnk";
        int? id, wholePercentages;
        IReadOnlyList<MetadataExtractor.Directory> directories;
        List<(string File, int WholePercentages)>? wholePercentagesCollection;
        bool fromDistanceContent = !file.EndsWith(lnk) && file.Contains(eDistanceContentDirectory);
        if (!file.EndsWith(lnk))
            (id, wholePercentages) = IMapping.GetConverted(configuration.FacesFileNameExtension, file);
        else
            (id, wholePercentages) = IMapping.GetConverted(configuration.FacesFileNameExtension, file[..^4]);
        if (id is null || wholePercentages is null)
            return;
        if (skipCollection.TryGetValue(id.Value, out wholePercentagesCollection))
        {
            fileMatches = (from l in wholePercentagesCollection where l.WholePercentages == wholePercentages select l.File).ToArray();
            foreach (string fileMatch in fileMatches)
            {
                if (string.IsNullOrEmpty(fileMatch) || !File.Exists(fileMatch))
                    continue;
                if (!fileMatch.EndsWith(".dup") && !File.Exists($"{fileMatch}.dup"))
                    File.Move(fileMatch, $"{fileMatch}.dup");
            }
        }
        if (file.EndsWith(lnk) || (!configuration.DistanceMoveUnableToMatch && !configuration.DistanceRenameToMatch) || !File.Exists(file))
            directories = new List<MetadataExtractor.Directory>();
        else
            directories = MetadataExtractor.ImageMetadataReader.ReadMetadata(file);
        RectangleF? rectangle = ILocation.GetPercentagesRectangle(configuration.LocationDigits, wholePercentages.Value);
        lock (locationContainers)
            locationContainers.Add(new(fromDistanceContent, directoryNumber, file, personKey, id.Value, wholePercentages.Value, directories, rectangle, null));
    }

    private static void OpenPossibleDuplicates(Configuration configuration, List<(long, int, string, float?)> duplicates)
    {
        string personKeyFormatted;
        foreach ((long personKey, int id, string file, float? percent) in duplicates)
        {
            if (percent is null)
                continue;
            _ = Process.Start("explorer.exe", string.Concat("\"", Path.GetDirectoryName(file), "\""));
            personKeyFormatted = IPersonBirthday.GetFormatted(configuration.PersonBirthdayFormat, personKey);
        }
        foreach ((long personKey, int id, string file, float? percent) in duplicates)
        {
            if (percent is not null && percent.Value == 0)
                continue;
            _ = Process.Start("explorer.exe", string.Concat("\"", Path.GetDirectoryName(file), "\""));
            personKeyFormatted = IPersonBirthday.GetFormatted(configuration.PersonBirthdayFormat, personKey);
        }
        foreach ((long personKey, int id, string file, float? percent) in duplicates)
        {
            if (percent is not null && percent.Value > 0)
                continue;
            _ = Process.Start("explorer.exe", string.Concat("\"", Path.GetDirectoryName(file), "\""));
            personKeyFormatted = IPersonBirthday.GetFormatted(configuration.PersonBirthdayFormat, personKey);
        }
    }

    private static void LookForPossibleDuplicates(Configuration configuration, ReadOnlyCollection<LocationContainer<MetadataExtractor.Directory>> locationContainers)
    {
        string key;
        float? percent;
        float itemPercentagesArea;
        List<string> delete = new();
        RectangleF? itemPercentagesRectangle;
        (string File, int WholePercentages) item;
        Dictionary<string, (string, int)> distinct = new();
        List<(long, int, string, float?)> duplicates = new();
        foreach (LocationContainer<MetadataExtractor.Directory> locationContainer in locationContainers)
        {
            key = string.Concat(locationContainer.PersonKey, locationContainer.Id);
            if (distinct.TryGetValue(key, out item))
            {
                if (item.WholePercentages == locationContainer.WholePercentages)
                    continue;
                itemPercentagesRectangle = ILocation.GetPercentagesRectangle(configuration.LocationDigits, item.WholePercentages);
                if (itemPercentagesRectangle is null || locationContainer.Rectangle is null)
                    percent = null;
                else
                {
                    itemPercentagesArea = itemPercentagesRectangle.Value.Width * itemPercentagesRectangle.Value.Height;
                    percent = ILocation.GetIntersectPercent(itemPercentagesRectangle.Value, itemPercentagesArea, locationContainer.Rectangle.Value);
                }
                delete.Add(item.File);
                delete.Add(locationContainer.File);
                duplicates.Add(new(locationContainer.PersonKey, locationContainer.Id, locationContainer.File, percent));
                continue;
            }
            distinct.Add(key, new(locationContainer.File, locationContainer.WholePercentages));
        }
        if (!configuration.DeletePossibleDuplicates && duplicates.Count > 0)
            OpenPossibleDuplicates(configuration, duplicates);
        else
        {
            if (delete.Count > 5)
                throw new Exception("Something maybe wrong!");
            foreach (string file in delete)
            {
                if (File.Exists(file))
                    File.Delete(file);
            }
        }
    }

    internal static List<LocationContainer<MetadataExtractor.Directory>> GetLocationContainers(Shared.Models.Methods.IDistance<MetadataExtractor.Directory> distance, int maxDegreeOfParallelism, Configuration configuration, long ticks, ReadOnlyCollection<PersonContainer> personContainers, string eDistanceContentDirectory, Dictionary<int, List<(string, int)>> skipCollection, List<Record> records)
    {
        List<LocationContainer<MetadataExtractor.Directory>> results = new();
        List<(long PersonKey, int? DirectoryNumber, string File)> collection = GetCollection(configuration, personContainers, records);
        if (collection.Count > 0 && (configuration.DistanceMoveUnableToMatch || configuration.DistanceRenameToMatch))
        {
            int totalSeconds = (int)Math.Floor(new TimeSpan(DateTime.Now.Ticks - ticks).TotalSeconds);
            string message = $") Building Mapped Face Files Collection - {totalSeconds} total second(s)";
            ParallelOptions parallelOptions = new() { MaxDegreeOfParallelism = maxDegreeOfParallelism };
            ProgressBarOptions options = new() { ProgressCharacter = '─', ProgressBarOnBottom = true, DisableBottomPercentage = true };
            using ProgressBar progressBar = new(collection.Count, message, options);
            _ = Parallel.For(0, collection.Count, parallelOptions, (i, state) =>
            {
                progressBar.Tick();
                ParallelFor(configuration, eDistanceContentDirectory, skipCollection, results, collection[i].PersonKey, collection[i].DirectoryNumber, collection[i].File);
            });
        }
        ReadOnlyCollection<LocationContainer<MetadataExtractor.Directory>> locationContainers = new(results.OrderBy(l => l.DirectoryNumber).ToArray());
        LookForPossibleDuplicates(configuration, locationContainers);
        if (configuration.LocationContainerDistanceTolerance is not null)
            distance.ReviewLocationContainerDistanceTolerance(configuration.LocationContainerDistanceTolerance.Value, locationContainers);
        return results;
    }

    internal static int CopyManualFiles(Configuration configuration, long ticks, ReadOnlyCollection<PersonContainer> personContainers, string eDistanceContentTicksDirectory)
    {
        int result = 0;
        string fileName;
        string checkFile;
        string? directory;
        string dateDirectory;
        string directoryName;
        string checkDirectory;
        string personKeyFormatted;
        PersonBirthday personBirthday;
        List<string> distinct = new();
        DateTime dateTime = new(ticks);
        string by = nameof(Shared.Models.Stateless.IMapLogic.ManualCopy);
        foreach (PersonContainer personContainer in personContainers)
        {
            if (personContainer.Key is null)
                continue;
            for (int i = personContainer.DisplayDirectoryAllFiles.Length - 1; i > -1; i--)
            {
                if (!personContainer.DisplayDirectoryAllFiles[i].EndsWith(configuration.FacesFileNameExtension))
                    continue;
                fileName = Path.GetFileName(personContainer.DisplayDirectoryAllFiles[i]);
                if (distinct.Contains(fileName))
                    continue;
                distinct.Add(fileName);
                directory = Path.GetDirectoryName(personContainer.DisplayDirectoryAllFiles[i]);
                if (string.IsNullOrEmpty(directory))
                    continue;
                directoryName = Path.GetFileName(directory);
                if (directoryName != personContainer.DisplayDirectoryName)
                    continue;
                personBirthday = IPersonBirthday.GetPersonBirthday(personContainer.Key.Value);
                personKeyFormatted = IPersonBirthday.GetFormatted(configuration.PersonBirthdayFormat, personBirthday);
                dateDirectory = Path.Combine(eDistanceContentTicksDirectory, by, personKeyFormatted, dateTime.ToString("yyyy"));
                checkDirectory = Path.Combine(dateDirectory, personContainer.DisplayDirectoryName);
                if (!Directory.Exists(checkDirectory))
                    _ = Directory.CreateDirectory(checkDirectory);
                checkFile = Path.Combine(dateDirectory, fileName);
                if (File.Exists(checkFile))
                    continue;
                File.Move(personContainer.DisplayDirectoryAllFiles[i], checkFile);
                personContainer.DisplayDirectoryAllFiles[i] = string.Empty;
                result++;
            }
        }
        return result;
    }

    private static string GetMappingSegmentB(long ticks, PersonBirthday personBirthday, int? approximateYears, long dateTimeOriginalThenMinimumDateTimeTicks, bool? isWrongYear)
    {
        int years;
        string result;
        TimeSpan? timeSpan = IPersonBirthday.GetTimeSpan(dateTimeOriginalThenMinimumDateTimeTicks, isWrongYear, personBirthday);
        if (timeSpan.HasValue && timeSpan.Value.Ticks < 0)
            result = "!---";
        else if (timeSpan.HasValue)
        {
            (years, _) = IPersonBirthday.GetAge(dateTimeOriginalThenMinimumDateTimeTicks, personBirthday);
            result = $"^{years:000}";
        }
        else if (approximateYears.HasValue)
        {
            DateTime dateTime = new(ticks);
            (years, _) = IAge.GetAge(dateTimeOriginalThenMinimumDateTimeTicks, dateTime.AddYears(-approximateYears.Value));
            result = $"~{years:000}";
        }
        else
        {
            string isWrongYearFlag = IItem.GetWrongYearFlag(isWrongYear);
            result = $"{isWrongYearFlag}{new DateTime(dateTimeOriginalThenMinimumDateTimeTicks):yyyy}";
        }
        return result;
    }

    internal static string GetMappingSegmentB(long ticks, PersonBirthday personBirthday, int? approximateYears, DateTime dateTimeOriginalThenMinimumDateTime, bool? isWrongYear)
    {
        string result = GetMappingSegmentB(ticks, personBirthday, approximateYears, dateTimeOriginalThenMinimumDateTime.Ticks, isWrongYear);
        return result;
    }

    internal static string GetMappingSegmentB(long ticks, PersonBirthday personBirthday, int? approximateYears, MappingFromItem mappingFromItem)
    {
        string result = GetMappingSegmentB(ticks, personBirthday, approximateYears, mappingFromItem.GetDateTimeOriginalThenMinimumDateTime(), mappingFromItem.IsWrongYear);
        return result;
    }

    internal static string? GetFacePartsContentCollectionFile(string extension, string d2FacePartsContentCollectionDirectory, MappingFromItem mappingFromItem)
    {
        string? result;
        string? directoryName = Path.GetDirectoryName(mappingFromItem.RelativePath);
        if (directoryName is null)
            result = null;
        else
            result = Path.Combine($"{d2FacePartsContentCollectionDirectory}{directoryName}", $"{mappingFromItem.ImageFileHolder.Name}{extension}");
        return result;
    }

    internal static string GetFacesDirectory(Shared.Models.Properties.IPropertyConfiguration propertyConfiguration, string dFacesContentDirectory, MappingFromItem mappingFromItem)
    {
        string result;
        (string directoryName, _) = IPath.GetDirectoryNameAndIndex(propertyConfiguration.ResultAllInOneSubdirectoryLength, mappingFromItem.ImageFileHolder.NameWithoutExtension);
        result = Path.Combine(dFacesContentDirectory, propertyConfiguration.ResultAllInOne, directoryName, mappingFromItem.ImageFileHolder.NameWithoutExtension);
        return result;
    }

    internal static string GetFacePartsDirectory(Shared.Models.Properties.IPropertyConfiguration propertyConfiguration, string d2FacePartsContentDirectory, MappingFromItem mappingFromItem)
    {
        string result;
        (string directoryName, _) = IPath.GetDirectoryNameAndIndex(propertyConfiguration.ResultAllInOneSubdirectoryLength, mappingFromItem.ImageFileHolder.NameWithoutExtension);
        result = Path.Combine(d2FacePartsContentDirectory, propertyConfiguration.ResultAllInOne, directoryName, mappingFromItem.ImageFileHolder.NameWithoutExtension);
        return result;
    }

    internal static SaveContainer GetDebugSaveContainer(string directory, MappingFromPerson? mappingFromPerson, SortingContainer sortingContainer, Mapping mapping)
    {
        SaveContainer result;
        if (sortingContainer.Mapping.MappingFromLocation is null)
            throw new NullReferenceException(nameof(sortingContainer.Mapping.MappingFromLocation));
        FileHolder faceFileHolder = new($"C:/{sortingContainer.Sorting.Id}.{sortingContainer.Sorting.WholePercentages}");
        string shortcutFile;
        if (mappingFromPerson is null)
            shortcutFile = Path.Combine(directory, $"{sortingContainer.Mapping.MappingFromLocation.DeterministicHashCodeKey}{sortingContainer.Mapping.MappingFromItem.ImageFileHolder.ExtensionLowered}.{sortingContainer.Sorting.DistancePermyriad}.lnk");
        else
            shortcutFile = Path.Combine(directory, $"{sortingContainer.Mapping.MappingFromLocation.DeterministicHashCodeKey}{sortingContainer.Mapping.MappingFromItem.ImageFileHolder.ExtensionLowered}.{sortingContainer.Sorting.DistancePermyriad}.lnk");
        result = new(directory, faceFileHolder, mapping.MappingFromItem.ResizedFileHolder, shortcutFile);
        return result;
    }

    internal static (SaveContainer?, SaveContainer?) GetContainers(string facesFileNameExtension, string facePartsFileNameExtension, Shared.Models.Properties.IPropertyConfiguration propertyConfiguration, string dFacesContentDirectory, string d2FacePartsContentCollectionDirectory, string directory, SortingContainer sortingContainer, Mapping mapping)
    {
        SaveContainer? result;
        SaveContainer? saveContainer;
        if (sortingContainer.Mapping.MappingFromLocation is null)
            throw new NullReferenceException(nameof(sortingContainer.Mapping.MappingFromLocation));
        if (mapping.MappingFromLocation is null)
            (result, saveContainer) = (null, null);
        else
        {
            string? facePartsContentCollectionFile = GetFacePartsContentCollectionFile(facePartsFileNameExtension, d2FacePartsContentCollectionDirectory, mapping.MappingFromItem);
            if (facePartsContentCollectionFile is null || !File.Exists(facePartsContentCollectionFile))
                result = null;
            else
            {
                string checkFile = Path.Combine(directory, $"{mapping.MappingFromItem.ImageFileHolder.Name}{facePartsFileNameExtension}");
                result = new(checkFile, directory, new(facePartsContentCollectionFile));
            }
            string facesDirectory = GetFacesDirectory(propertyConfiguration, dFacesContentDirectory, mapping.MappingFromItem);
            FileHolder faceFileHolder = new(Path.Combine(facesDirectory, $"{mapping.MappingFromLocation.DeterministicHashCodeKey}{mapping.MappingFromItem.ImageFileHolder.ExtensionLowered}{facesFileNameExtension}"));
            if (!faceFileHolder.Exists)
                saveContainer = null;
            else
            {
                string checkFile = Path.Combine(directory, $"{mapping.MappingFromLocation.DeterministicHashCodeKey}{mapping.MappingFromItem.ImageFileHolder.ExtensionLowered}{facesFileNameExtension}");
                saveContainer = new(checkFile, directory, faceFileHolder);
            }
        }
        return (result, saveContainer);
    }

    internal static (SaveContainer, SaveContainer?) GetContainers(string facesFileNameExtension, string facePartsFileNameExtension, string directory, FileHolder faceFileHolder, FileHolder facePartsFileHolder, Mapping mapping)
    {
        string checkFile;
        SaveContainer? saveContainer;
        if (mapping.MappingFromLocation is null)
            throw new NullReferenceException(nameof(mapping.MappingFromLocation));
        checkFile = Path.Combine(directory, $"{mapping.MappingFromItem.ImageFileHolder.Name}{facePartsFileNameExtension}");
        saveContainer = !facePartsFileHolder.Exists ? null : new(checkFile, directory, facePartsFileHolder);
        checkFile = Path.Combine(directory, $"{mapping.MappingFromLocation.DeterministicHashCodeKey}{mapping.MappingFromItem.ImageFileHolder.ExtensionLowered}{facesFileNameExtension}");
        return (new(checkFile, directory, faceFileHolder), saveContainer);
    }

    private static IEnumerable<(string, string)> GetCollection(string[] yearDirectories)
    {
        foreach (string l in yearDirectories)
            yield return new(l, Path.GetFileName(l));
    }

    internal static void SaveMappingShortcuts(string mappingDirectory)
    {
        string? shortcutFileName;
        string[] yearDirectories;
        string personKeyFormatted;
        string[] personNameDirectories;
        WindowsShortcut windowsShortcut;
        string personDisplayDirectoryName;
        (string, string)[] yearDirectoryNameCheck;
        List<(string, string)> yearDirectoryNames = new();
        string[] personKeyDirectories = Directory.GetDirectories(mappingDirectory, "*", SearchOption.TopDirectoryOnly);
        foreach (string personKeyDirectory in personKeyDirectories)
        {
            windowsShortcut = new();
            shortcutFileName = null;
            yearDirectoryNames.Clear();
            personKeyFormatted = Path.GetFileName(personKeyDirectory);
            yearDirectories = Directory.GetDirectories(personKeyDirectory, "*", SearchOption.TopDirectoryOnly);
            yearDirectoryNames.AddRange(GetCollection(yearDirectories));
            yearDirectoryNameCheck = (from l in yearDirectoryNames where l.Item2.Contains('^') select l).OrderByDescending(l => l.Item2).ToArray();
            if (yearDirectoryNameCheck.Length == 0)
                yearDirectoryNameCheck = (from l in yearDirectoryNames where l.Item2.Contains('~') select l).OrderByDescending(l => l.Item2).ToArray();
            if (yearDirectoryNameCheck.Length == 0)
                yearDirectoryNameCheck = (from l in yearDirectoryNames where l.Item2.Contains('=') select l).OrderByDescending(l => l.Item2).ToArray();
            if (yearDirectoryNameCheck.Length == 0)
                yearDirectoryNameCheck = (from l in yearDirectoryNames select l).OrderByDescending(l => l).ToArray();
            if (yearDirectoryNameCheck.Length == 0)
                continue;
            foreach ((string yearDirectory, string yearDirectoryName) in yearDirectoryNameCheck)
            {
                personNameDirectories = Directory.GetDirectories(yearDirectory, "*", SearchOption.TopDirectoryOnly);
                foreach (string personNameDirectory in personNameDirectories)
                {
                    personDisplayDirectoryName = Path.GetFileName(personNameDirectory).Split('-')[0];
                    if (personDisplayDirectoryName is null)
                        continue;
                    windowsShortcut.Path = yearDirectory;
                    windowsShortcut.Description = yearDirectoryName;
                    shortcutFileName = Path.Combine(mappingDirectory, $"{personDisplayDirectoryName} [{windowsShortcut.Description}].lnk");
                    break;
                }
                if (shortcutFileName is not null)
                {
                    if (!File.Exists(shortcutFileName))
                        break;
                }
            }
            if (shortcutFileName is null || windowsShortcut.Path is null || windowsShortcut.Description is null)
                continue;
            try
            {
                windowsShortcut.Save(shortcutFileName);
                windowsShortcut.Dispose();
            }
            catch (Exception)
            { }
        }
    }

    internal static Dictionary<int, List<long>> GetIdToPersonKeys(Dictionary<long, List<int>> personKeyToIds)
    {
        Dictionary<int, List<long>> results = new();
        List<long>? collection;
        foreach (KeyValuePair<long, List<int>> keyValuePair in personKeyToIds)
        {
            foreach (int id in keyValuePair.Value)
            {
                if (!results.TryGetValue(id, out collection))
                {
                    results.Add(id, new());
                    if (!results.TryGetValue(id, out collection))
                        throw new Exception();
                }
                if (collection.Contains(keyValuePair.Key))
                    continue;
                collection.Add(keyValuePair.Key);
            }
        }
        return results;
    }

    internal static Mapping[] GetSelectedMappingCollection(ReadOnlyCollection<Face> faces)
    {
        Mapping[] results;
        IEnumerable<Mapping> collection = from l in faces orderby l.Mapping?.MappingFromItem.Id select l.Mapping;
        results = (from l in collection where l is not null select l).ToArray();
        return results;
    }

    internal static ReadOnlyCollection<Face> GetFaces(ReadOnlyCollection<Item> items)
    {
        List<Face> results = new();
        foreach (Item item in items)
        {
            if (item.Property?.Id is null || item.ResizedFileHolder is null)
                continue;
            foreach (Face face in item.Faces)
            {
                if (face.FaceEncoding is null || face.Location is null || face.OutputResolution is null)
                    continue;
                results.Add(face);
            }
        }
        return new(results);
    }

    internal static Mapping[] GetSelectedMappingCollection(ReadOnlyCollection<Item> items)
    {
        Mapping[] results;
        ReadOnlyCollection<Face> faces = GetFaces(items);
        results = GetSelectedMappingCollection(faces);
        return results;
    }

    private static ReadOnlyDictionary<int, ReadOnlyDictionary<int, Mapping>> GetReadOnly(Dictionary<int, Dictionary<int, Mapping>> keyValuePairs)
    {
        Dictionary<int, ReadOnlyDictionary<int, Mapping>> results = new();
        foreach (KeyValuePair<int, Dictionary<int, Mapping>> keyValuePair in keyValuePairs)
            results.Add(keyValuePair.Key, new(keyValuePair.Value));
        return new(results);
    }

    internal static ReadOnlyDictionary<int, ReadOnlyDictionary<int, Mapping>> GetIdToWholePercentagesToFace(ReadOnlyCollection<Mapping> mappingCollection)
    {
        Dictionary<int, Dictionary<int, Mapping>> results = new();
        Dictionary<int, Mapping>? keyValuePairs;
        foreach (Mapping mapping in mappingCollection)
        {
            if (mapping.MappingFromLocation is null)
                continue;
            if (!results.TryGetValue(mapping.MappingFromItem.Id, out keyValuePairs))
            {
                results.Add(mapping.MappingFromItem.Id, new());
                if (!results.TryGetValue(mapping.MappingFromItem.Id, out keyValuePairs))
                    throw new Exception();
            }
            if (keyValuePairs.ContainsKey(mapping.MappingFromLocation.WholePercentages))
                continue;
            keyValuePairs.Add(mapping.MappingFromLocation.WholePercentages, mapping);
        }
        return GetReadOnly(results);
    }

    private static string? TryToFind(char[] personCharacters, string a2PeopleSingletonDirectory, List<(string Directory, string DirectoryName, string DirectoryNameSplitFirst)> a2PeopleSingletonDirectories, string file, string path)
    {
        string? result;
        string? pathName = Path.GetFileName(path);
        string? group = Path.GetDirectoryName(path);
        string? groupName = Path.GetFileName(group);
        if (pathName is null || group is null || groupName is null)
            result = null;
        else
        {
            string[] matches;
            matches = (from l in a2PeopleSingletonDirectories where l.DirectoryName == pathName select l.Directory).ToArray();
            if (matches.Length == 1)
                result = matches.First();
            else
            {
                string pathNameSplitFirst = pathName.Split(personCharacters).First();
                matches = (from l in a2PeopleSingletonDirectories where l.DirectoryNameSplitFirst == pathNameSplitFirst select l.Directory).ToArray();
                if (matches.Length == 1)
                    result = matches.First();
                else
                {
                    string checkDirectory = Path.Combine(a2PeopleSingletonDirectory, groupName, pathName);
                    if (!Directory.Exists(checkDirectory))
                        result = null;
                    else
                        result = checkDirectory;
                }
            }
            if (!string.IsNullOrEmpty(result))
            {
                try
                {
                    WindowsShortcut windowsShortcut;
                    windowsShortcut = new() { Path = result };
                    windowsShortcut.Save(file);
                    windowsShortcut.Dispose();
                }
                catch (Exception)
                { }
            }
        }
        return result;
    }

    private static List<(string, long)> GetJLinkResolvedDirectories(string personBirthdayFormat, List<string> resolvedDirectories)
    {
        List<(string, long)> results = new();
        DateTime dateTime;
        string directoryName;
        string[] directories;
        foreach (string resolvedDirectory in resolvedDirectories)
        {
            directories = Directory.GetDirectories(resolvedDirectory, "*", SearchOption.TopDirectoryOnly);
            foreach (string directory in directories)
            {
                directoryName = Path.GetFileName(directory);
                if (directoryName.Length != personBirthdayFormat.Length || !DateTime.TryParseExact(directoryName, personBirthdayFormat, CultureInfo.InvariantCulture, DateTimeStyles.None, out dateTime))
                    continue;
                results.Add((resolvedDirectory, dateTime.Ticks));
            }
        }
        if (results.Count != resolvedDirectories.Count)
            results.Clear();
        return results;
    }

    private static List<(string, long)> GetGenealogicalDataCommunicationDirectories(string genealogicalDataCommunicationFile, string[] jLinks, string personBirthdayFormat)
    {
        List<(string, long)> results = new();
        string directory;
        DateTime dateTime;
        string[] personKeyDirectories;
        string[] personDisplayDirectoryNames;
        string personKeyFormattedDirectoryName;
        string? genealogicalDataCommunicationDirectory = Path.GetDirectoryName(genealogicalDataCommunicationFile);
        foreach (string jLink in jLinks)
        {
            if (genealogicalDataCommunicationDirectory is null)
                continue;
            directory = Path.Combine(genealogicalDataCommunicationDirectory, jLink);
            if (!Directory.Exists(directory))
                continue;
            personDisplayDirectoryNames = Directory.GetDirectories(directory, "*", SearchOption.TopDirectoryOnly);
            foreach (string personDisplayDirectoryName in personDisplayDirectoryNames)
            {
                personKeyDirectories = Directory.GetDirectories(personDisplayDirectoryName, "*", SearchOption.TopDirectoryOnly);
                foreach (string personKeyFormattedDirectory in personKeyDirectories)
                {
                    personKeyFormattedDirectoryName = Path.GetFileName(personKeyFormattedDirectory);
                    if (personKeyFormattedDirectoryName.Length != personBirthdayFormat.Length || !DateTime.TryParseExact(personKeyFormattedDirectoryName, personBirthdayFormat, CultureInfo.InvariantCulture, DateTimeStyles.None, out dateTime))
                        continue;
                    results.Add((directory, dateTime.Ticks));
                }
            }
        }
        return results;
    }

    internal static List<(string, long)> GetJLinkDirectories(string genealogicalDataCommunicationFile, string[] jLinks, string personBirthdayFormat, char[] personCharacters, string a2PeopleSingletonDirectory, string a2PeopleContentDirectory)
    {
        List<(string, long)> results;
        string[] files;
        string? foundPath;
        int totalFiles = 0;
        string checkDirectory;
        WindowsShortcut windowsShortcut;
        List<string> resolvedDirectories = new();
        results = GetGenealogicalDataCommunicationDirectories(genealogicalDataCommunicationFile, jLinks, personBirthdayFormat);
        if (results.Count == 0 || results.Count < jLinks.Length)
        {
            List<(string, string, string)> a2PeopleSingletonDirectories = new();
            foreach (string directory in Directory.GetDirectories(a2PeopleSingletonDirectory, "*", SearchOption.AllDirectories))
                a2PeopleSingletonDirectories.Add((directory, Path.GetFileName(directory), Path.GetFileName(directory).Split(personCharacters).First()));
            foreach (string directoryName in jLinks)
            {
                checkDirectory = Path.Combine(a2PeopleContentDirectory, directoryName);
                if (!Directory.Exists(checkDirectory))
                    continue;
                files = Directory.GetFiles(checkDirectory, "*.lnk", SearchOption.TopDirectoryOnly);
                totalFiles += files.Length;
                foreach (string file in files)
                {
                    windowsShortcut = WindowsShortcut.Load(file);
                    if (windowsShortcut.Path is null)
                        continue;
                    if (Directory.Exists(windowsShortcut.Path))
                        resolvedDirectories.Add(windowsShortcut.Path);
                    else
                    {
                        foundPath = TryToFind(personCharacters, a2PeopleSingletonDirectory, a2PeopleSingletonDirectories, file, windowsShortcut.Path);
                        if (string.IsNullOrEmpty(foundPath))
                            continue;
                        resolvedDirectories.Add(foundPath);
                    }
                }
            }
            if (totalFiles == resolvedDirectories.Count)
                results = GetJLinkResolvedDirectories(personBirthdayFormat, resolvedDirectories);
            else
            {
                resolvedDirectories.Clear();
                results = new();
            }
        }
        return results;
    }

    internal static (string, bool, bool) Get(int? useFiltersCounter, bool saveIndividually, bool sortingContainersAny, string forceSingleImageHumanized, Mapping mapping)
    {
        string by;
        bool isByMapping;
        bool isBySorting;
        if (mapping.By is null)
        {
            isByMapping = false;
            isBySorting = !sortingContainersAny;
            by = $"{nameof(Shared.Models.Stateless.IMapLogic.Mapping)}Null";
        }
        else
        {
            isByMapping = mapping.By == Shared.Models.Stateless.IMapLogic.Mapping;
            isBySorting = mapping.By == Shared.Models.Stateless.IMapLogic.Sorting;
            if (isBySorting && mapping.MappingFromPerson is null)
                by = saveIndividually ? nameof(Shared.Models.Stateless.IMapLogic.Individually) : $"{nameof(Shared.Models.Stateless.IMapLogic.Sorting)} Without Person";
            else if (isBySorting && useFiltersCounter.HasValue)
                by = $"{nameof(Shared.Models.Stateless.IMapLogic.Sorting)} Modified Filters - {useFiltersCounter.Value}";
            else
            {
                by = mapping.By.Value switch
                {
                    Shared.Models.Stateless.IMapLogic.Mapping => nameof(Shared.Models.Stateless.IMapLogic.Mapping),
                    Shared.Models.Stateless.IMapLogic.Sorting => saveIndividually ? nameof(Shared.Models.Stateless.IMapLogic.Individually) : nameof(Shared.Models.Stateless.IMapLogic.Sorting),
                    Shared.Models.Stateless.IMapLogic.ForceSingleImage => forceSingleImageHumanized,
                    _ => throw new NotImplementedException()
                };
            }
        }
        return new(by, isByMapping, isBySorting);
    }

    internal static void CheckCollection(Shared.Models.Properties.IPropertyConfiguration propertyConfiguration, string? rootDirectoryParent)
    {
        string json;
        string fullPath;
        List<KeyValuePair<int, int[]>>? collection;
        foreach (string propertyContentCollectionFile in propertyConfiguration.PropertyContentCollectionFiles)
        {
            fullPath = Path.GetFullPath(string.Concat(rootDirectoryParent, propertyContentCollectionFile));
            if (fullPath.Contains(propertyConfiguration.RootDirectory))
                continue;
            if (!File.Exists(fullPath))
                continue;
            json = File.ReadAllText(fullPath);
            collection = JsonSerializer.Deserialize<List<KeyValuePair<int, int[]>>>(json);
            if (collection is null)
                throw new NullReferenceException(nameof(collection));
        }
    }

    internal static void LookForAbandoned(List<int> distinctFilteredIds, string directory, string directoryName)
    {
        string fileNameWithoutExtension;
        bool nameWithoutExtensionIsIdFormat;
        List<string> renameCollection = new();
        bool nameWithoutExtensionIsPaddedIdFormat;
        int sortOrderOnlyLengthIndex = IDirectory.GetSortOrderOnlyLengthIndex();
        string[] distinctFilteredIdsValues = distinctFilteredIds.Select(l => l.ToString()).ToArray();
        string[] files = Directory.GetFiles(directory, "*", SearchOption.AllDirectories);
        foreach (string file in files)
        {
            fileNameWithoutExtension = Path.GetFileNameWithoutExtension(Path.GetFileNameWithoutExtension(Path.GetFileNameWithoutExtension(file)));
            nameWithoutExtensionIsIdFormat = IProperty.NameWithoutExtensionIsIdFormat(fileNameWithoutExtension);
            nameWithoutExtensionIsPaddedIdFormat = IDirectory.NameWithoutExtensionIsPaddedIdFormat(fileNameWithoutExtension, sortOrderOnlyLengthIndex);
            if (!nameWithoutExtensionIsIdFormat && !nameWithoutExtensionIsPaddedIdFormat)
                continue;
            if (distinctFilteredIdsValues.Contains(fileNameWithoutExtension))
                continue;
            renameCollection.Add(file);
        }
        if (renameCollection.Count > 0)
        {
            if (directoryName.Length == 2)
                IDirectory.MoveFiles(renameCollection, directoryName, $"{directoryName[0]}abd{directoryName[^1]}");
            else if (directoryName.Length == 4)
                IDirectory.MoveFiles(renameCollection, directoryName, $"{directoryName[..2]}abd{directoryName[^2..]}");
            else
                throw new NotSupportedException();
        }
    }

    internal static void LookForAbandoned(string bResultsFullGroupDirectory, List<int> distinctFilteredIds)
    {
        string[] directories = Directory.GetDirectories(bResultsFullGroupDirectory, "*", SearchOption.TopDirectoryOnly);
        foreach (string directory in directories)
        {
            string? directoryName = Path.GetFileName(directory);
            if (string.IsNullOrEmpty(directoryName) || (directoryName.Length != 2 && directoryName.Length != 4))
                continue;
            LookForAbandoned(distinctFilteredIds, directory, directoryName);
        }
    }

    internal static ReadOnlyDictionary<int, List<int>> ConvertSkip(Dictionary<int, List<(string, int)>> skipCollection)
    {
        Dictionary<int, List<int>> results = new();
        List<int>? wholePercentagesCollection;
        foreach (KeyValuePair<int, List<(string, int)>> keyValuePair in skipCollection)
        {
            if (!results.TryGetValue(keyValuePair.Key, out wholePercentagesCollection))
            {
                results.Add(keyValuePair.Key, new());
                if (!results.TryGetValue(keyValuePair.Key, out wholePercentagesCollection))
                    throw new Exception();
            }
            foreach ((string _, int wholePercentage) in keyValuePair.Value)
                wholePercentagesCollection.Add(wholePercentage);
        }
        return new(results);
    }

    internal static ReadOnlyDictionary<int, List<int>> ConvertSkipNotSkip(Dictionary<int, List<(string, int)>> skipNotSkipCollection)
    {
        Dictionary<int, List<int>> results = new();
        List<int>? wholePercentagesCollection;
        foreach (KeyValuePair<int, List<(string, int)>> keyValuePair in skipNotSkipCollection)
        {
            if (!results.TryGetValue(keyValuePair.Key, out wholePercentagesCollection))
            {
                results.Add(keyValuePair.Key, new());
                if (!results.TryGetValue(keyValuePair.Key, out wholePercentagesCollection))
                    throw new Exception();
            }
            foreach ((string _, int wholePercentage) in keyValuePair.Value)
                wholePercentagesCollection.Add(wholePercentage);
        }
        return new(results);
    }

    internal static ReadOnlyDictionary<int, List<LocationContainer<MetadataExtractor.Directory>>> ConvertLocationContainers(List<LocationContainer<MetadataExtractor.Directory>> locationContainers)
    {
        Dictionary<int, List<LocationContainer<MetadataExtractor.Directory>>> results = new();
        foreach (LocationContainer<MetadataExtractor.Directory> locationContainer in locationContainers)
        {
            if (!results.ContainsKey(locationContainer.Id))
                results.Add(locationContainer.Id, new());
            results[locationContainer.Id].Add(locationContainer);
        }
        return new(results);
    }

    internal static void MoveToDecade(Property.Models.Configuration propertyConfiguration, MappingFromItem mappingFromItem, MappingFromPerson mappingFromPerson)
    {
        string year;
        DateTime dateTime;
        FileInfo fileInfo;
        string halfDecade;
        string checkDirectory;
        string? yearDirectory;
        string yearDirectoryName;
        string? personNameDirectory;
        string personNameDirectoryName;
        string? personKeyFormattedDirectory;
        string? personKeyFormattedDirectoryName;
        foreach (string locationContainersFile in mappingFromPerson.LocationContainersFiles)
        {
            fileInfo = new(locationContainersFile);
            if (!fileInfo.Exists)
                continue;
            personNameDirectory = Path.GetDirectoryName(locationContainersFile);
            if (string.IsNullOrEmpty(personNameDirectory))
                continue;
            personNameDirectoryName = Path.GetFileName(personNameDirectory);
            yearDirectory = Path.GetDirectoryName(personNameDirectory);
            if (string.IsNullOrEmpty(yearDirectory))
                continue;
            yearDirectoryName = Path.GetFileName(yearDirectory);
            personKeyFormattedDirectory = Path.GetDirectoryName(yearDirectory);
            if (string.IsNullOrEmpty(personKeyFormattedDirectory))
                continue;
            personKeyFormattedDirectoryName = Path.GetFileName(personKeyFormattedDirectory);
            if (personKeyFormattedDirectoryName.Length != propertyConfiguration.PersonBirthdayFormat.Length)
                break;
            if (mappingFromItem.DateTimeOriginal is null)
            {
                dateTime = mappingFromItem.MinimumDateTime;
                year = mappingFromItem.MinimumDateTime.Year.ToString();
                halfDecade = year[3] > '4' ? $"#{year[..3]}+" : $"#{year[..3]}-";
            }
            else
            {
                dateTime = mappingFromItem.DateTimeOriginal.Value;
                year = mappingFromItem.DateTimeOriginal.Value.Year.ToString();
                halfDecade = year[3] > '4' ? $"^{year[..3]}+" : $"^{year[..3]}-";
            }
            if (fileInfo.CreationTime != dateTime)
                File.SetCreationTime(locationContainersFile, dateTime);
            if (halfDecade == yearDirectoryName)
                continue;
            checkDirectory = Path.Combine(personKeyFormattedDirectory, halfDecade, personNameDirectoryName);
            if (!Directory.Exists(checkDirectory))
                _ = Directory.CreateDirectory(checkDirectory);
            File.Move(locationContainersFile, Path.Combine(checkDirectory, Path.Combine(checkDirectory, Path.GetFileName(locationContainersFile))));
        }
    }

}